#!/usr/bin/env python

# API Header parameter
header_param_name = 'project_id'

# Blockfrost Network URL
bf_url_cardano_mainnet = 'https://cardano-mainnet.blockfrost.io/api/v0/'
bf_url_cardano_preprod = 'https://cardano-preprod.blockfrost.io/api/v0/'
bf_url_cardano_preview = 'https://cardano-preview.blockfrost.io/api/v0/'
bf_url_cardano_legacy  = 'https://cardano-testnet.blockfrost.io/api/v0/' #testnet

# Blockfrost Stake URL
bf_stake_url = 'accounts/'
bf_stake_rewards_url = '/rewards'
bf_stake_amount_history_url = '/history'
bf_stake_delegation_url = '/delegations'
bf_stake_registration_url = '/registrations'
bf_stake_withdrawal_history_url = '/withdrawals'
bf_stake_mir_history_url = '/mirs'
bf_associated_addresses_url = '/addresses'
bf_assets_associated_addresses_url = '/addresses/assets'

# Blockfrost Addresse URL
bf_address_url = 'addresses/'
bf_address_details_url = '/total'
bf_address_utxo_url = '/utxos'
bf_address_transaction_url = '/transactions'

# Blockfrost Network Information URL
bf_network_informations_url = '/network'

# Blockfrost Epochs URL
bf_epoch_url = 'epochs/'
bf_latest_epoch_url = 'epochs/latest'
bf_latest_epoch_protocol_parameters_url = 'epochs/latest/parameters'

# Blockfrost Pools URL
bf_polls_url = 'pools/'
bf_param_stake_pool_history_url = '/history'

# Blockfrost Assets URL
bf_assets_url = 'assets/'
bf_asset_history_url = '/history'
bf_asset_transactions_url = '/transactions'
bf_asset_addresses_url = '/addresses'
bf_assets_policy_url = 'policy/'

# Blockfrost Transaction URL
bf_tx_url = 'txs/'
bf_tx_utxos_url = '/utxos'
bf_stake_address_certificates_url = '/stakes'
bf_tx_delegation_certificates_url = '/delegations'
bf_tx_withdrawal_url = '/withdrawals'
bf_tx_transaction_mirs_url = '/mirs'
bf_tx_stake_pool_update_url = '/pool_updates'
bf_tx_stake_pool_retirement_cert_url = '/pool_retires'
bf_tx_metadata_url = '/metadata'
bf_tx_cbor_metadata_url = '/metadata/cbor'
bf_tx_redeemers_url = '/redeemers'

# Blockfrost Script URL
bf_scripts_url = '/scripts'
bf_specific_script_url = '/scripts/'
bf_redeem_specific_script_url = '/redeemers'

# Blockfrost Blocks URL
bf_blocks_url = '/blocks'
bf_next_blocks_url = '/next'
bf_blocks_latest_url = '/latest'
bf_blocks_parameters_url = '/parameters'
bf_pevious_block_url = '/previous'
bf_blocks_slot_url = '/slot'
bf_blocks_epoch = '/epoch'
bf_blocks_tx_url = '/txs'
bf_blocks_addresses_url = '/addresses'

