bf_osenv_testnet_apiKey = "BLOCKFROST_TESTNET_API_KEY"
bf_osenv_mainnet_apiKey = "BLOCKFROST_MAINNET_API_KEY"
bf_osenv_preview_apiKey = "BLOCKFROST_PREVIEW_API_KEY"
bf_osenv_preprod_apiKey = "BLOCKFROST_PREPROD_API_KEY"

bf_osenv = {"testnet":bf_osenv_testnet_apiKey,
           "mainnet":bf_osenv_mainnet_apiKey,
           "preview":bf_osenv_preview_apiKey,
           "preprod":bf_osenv_preprod_apiKey} 
